#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <linux/tcp.h>

#ifndef	TCP_NODELAY
#define TCP_NODELAY	0x01
#endif

#include "rserial.h"

//
//	It's not necessary to collect the following variables 
//	into a structure.	-- William

int		ptylock[ MAXPORT ];
time_t		locktime[ MAXPORT ];
int		pfd[ MAXPORT ];		// pty fd
int		efd[ MAXPORT ];		// Ethernet fd
int		cfd[ MAXPORT ];		// connect fd
iopacket	iop[ MAXPORT ];	
int		replyflag[ MAXPORT ];

char		bigbuf[ MAXPORT ][ BIGBUFLEN ];

char		*ptr[ MAXPORT ];
int		rlen[ MAXPORT ];

char		genbuf[ BUFLEN + HEADER ];

int		iidx[ MAXPORT ];
int		ilock[ MAXPORT ];
time_t		itime[ MAXPORT ];
iopacket	iocbuf[ MAXPORT ];

int	ispacket( iopacket *iopp, char *end )
{
u32	bound;
	
	if ((u32)iopp >= (u32)end)
		return 0;


	bound = (u32)iopp->val.ptr + iopp->len;
	
	bound += 3;
	bound &= ~3;
	
	if (bound <= (u32)end)
		return 1;
	else
		return 0;
}


int	connect_serv( struct in_addr *ip, int port, int *fd )
{
int	sock;
int	flag;
struct	sockaddr_in in;
int	one = 1;

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		printf("Connot open socket in connect_serv\n");
		return -1;
	}

	flag = fcntl( sock, F_GETFL );
	flag |= O_NONBLOCK;
	fcntl( sock, F_SETFL, flag);
				
	in.sin_family = AF_INET;
	in.sin_port = htons( port );
	in.sin_addr.s_addr = ip->s_addr;

	if (connect(sock, (struct sockaddr *)&in, sizeof(struct sockaddr_in)) < 0) {
		printf("connect_serv: connect < 0\n");
		*fd = sock;
		return -1;
	} else
		printf("connect_serv: connection made.\n");
	
	if (setsockopt( sock, SOL_SOCKET, SO_KEEPALIVE, &one, sizeof one))
		printf("setsockopt SO_KEEPALIVE error %d \n", port);
		
	

	return sock;
}

void	send_cmd(int fd, int cmd, void *buf)
{
	((iopacket *)buf)->cmd = cmd;
	((iopacket *)buf)->magic = cmd;
	((iopacket *)buf)->len = 0;
	write( fd, buf, HEADER);
}

int main(int argc, char *argv[])
{
fd_set		rfds, wfds, efds;
iopacket	*iopp;
u32		n;
char		buf[100];

struct		sockaddr_in in;
struct		in_addr	ip;
struct		Rserial rs;
struct		timeval	tv;

int		i;
int		fd;
int		sock;
int		maxfd = 0;
int		ret, len;
int		numport;
int		one = 1;
int		val;

	if (argc < 2) {
		printf("No IP address assigned, use 127.0.0.1 for destination\n");
		argv[1] = "127.0.0.1";
	}
	if ( !inet_aton( argv[1], &ip) ) {
		printf("Invalid ip address : %s\n", argv[1]);
		exit(1);
	}

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if ( sock < 0) {
		printf("Cannot open sock\n");
		exit(1);
	}

	if ( setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, 
			(char *)&one, sizeof(one)) == -1) {
		printf("sockopt SO_REUSEADDR error\n");
		exit(1);
	}

	memset(&in, 0, sizeof( struct sockaddr_in) );
	
	in.sin_family = AF_INET;
	in.sin_port = htons(SERVER_PORT);
	in.sin_addr.s_addr = ip.s_addr;
	
	if (connect(sock, (struct sockaddr *)&in, sizeof(struct sockaddr_in)) < 0){
		printf("Cannot make connection to server\n");
		exit(1);
	}
	read( sock, &rs, sizeof(struct Rserial) );
	close(sock);
	
	printf("Read from Server\n");
	printf("Server has %d ports\n", rs.num);
	printf("The port number are :\n");
	for(i=0;i<rs.num;i++) 
		printf("\t%d\n", rs.port[i]);

	numport = rs.num;
	//
	//	Open ethernet 
	//
	for(i=0;i<numport;i++) {
		sock = socket(AF_INET, SOCK_STREAM, 0);
		if (sock < 0) {
			printf("Cannot open socket for %d\n", i);
			exit(1);
		}
		memset(&in, 0, sizeof(struct sockaddr_in) );
		
		in.sin_family = AF_INET;
		in.sin_port = htons( rs.port[i] );
		in.sin_addr.s_addr = ip.s_addr;
		
		if (connect(sock, (struct sockaddr *)&in, sizeof(struct sockaddr_in)) < 0) {
			printf("Cannot make connection for %d\n", i);
			exit(1);
		}
		n = 1;
		if (setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &n, sizeof(n)) < 0) {
			printf("setsockopt SO_KEEPALIVE failed\n");
			exit(1);
		}
		if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &n, sizeof(n)) < 0) {
			printf("setsockopt TCP_NODELAY failed\n");
			exit(1);
		}

		printf("Port %d connected\n", i);
		efd[i] = sock;
	}
	
	//
	//	Open pty
	//
	for(i=0;i<numport;i++) {
		sprintf(buf, "/dev/ptyrs%d", i);
		fd = open( buf, O_RDWR );
		if (fd < 0) {
			printf("Cannot open pty %d\n", i);
			exit(1);
		}
		pfd[i] = fd;
	}
	
	for(i=0;i<numport;i++) {
		ptr[i] = &bigbuf[i][0];
		rlen[i] = BIGBUFLEN;
		cfd[i] = -1;
	}
	memset( ptylock, 0, sizeof(ptylock));
	memset( replyflag, 0, sizeof(replyflag));
	memset( locktime, 0, sizeof(locktime));
	memset( iidx, 0, sizeof(iidx));
	memset( ilock, 0, sizeof(ilock));
	memset( itime, 0, sizeof(itime));
	memset( iocbuf, 0, sizeof(iocbuf));

	for(;;) {
		FD_ZERO( &rfds );
		FD_ZERO( &wfds );
		FD_ZERO( &efds );
		
		for(i=0;i<numport;i++) {
			if ( efd[i] != -1) {
				FD_SET( efd[i], &rfds );
				maxfd = (maxfd > efd[i]) ? maxfd : efd[i];
			}

			if ( ! ptylock[i] ) {
				FD_SET( pfd[i], &rfds );
				maxfd = (maxfd > pfd[i]) ? maxfd : pfd[i];
			}

			FD_SET( pfd[i], &efds );
			FD_SET( pfd[i], &wfds );

			if (cfd[i] != -1) {
				FD_SET( cfd[i], &wfds );
				maxfd = (maxfd > cfd[i]) ? maxfd : cfd[i];
			}
		}
		
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		n = select( maxfd + 1, &rfds, &wfds, &efds, &tv);
		
		if (n < 0) {
			printf("Select error !! errno = %d\n", errno);
			continue;
		}
		
		//
		//	CONFIRM
		//
		for(i=0;i<numport;i++) {
			if (locktime[i] && (( time(NULL)-locktime[i] )>CONFIRM_T) ) {
//printf("## CMD_CONFIRM\n");
				send_cmd( efd[i], CMD_CONFIRM, genbuf);
				time( &locktime[i] );
			}

			if (itime[i] && (( time(NULL)-itime[i] ) > CONFIRM_T) ) {

				// CMD_IOCTLACK not coming, resend ioctl cmd again
				
				write( efd[i], &iocbuf[i], IOCTL_BODY + HEADER );
				time( &itime[i] );
			}

		}

		if (n == 0) {


			continue;
		}

		//
		//	Reply with ACK
		//
		for(i=0;i<numport;i++) {
			if ( replyflag[i] && FD_ISSET( pfd[i], &wfds) ) {
				send_cmd( efd[i], CMD_ACK, genbuf);
				replyflag[i] = 0;
			}
		}


		for(i=0;i<numport;i++) {
			if ((cfd[i] != -1) && (FD_ISSET( cfd[i], &wfds))) {
				len = sizeof( val );
				if (getsockopt( cfd[i], SOL_SOCKET, SO_ERROR, &val, &len )) {
					printf("getsockopt error\n");
					exit(1);
				}
				if (val) {
					printf("connect error %d : %d \n", i, val);
					close( cfd[i] );
					iidx[i] = 0;
					efd[i] = connect_serv( &ip, rs.port[i], &cfd[i] );
					continue;
				} else {
					printf("Re-connect to server ok %d\n", i);
					efd[i] = cfd[i];
					cfd[i] = -1;
				}
			}
		}

		//
		//	Datum come from Ethernet
		//
		//	1. Data	 (cmd == -1)
		//	2. Ioctl (cmd != -1)
		//	3. ACK.  (cmd == -1) && (len == 0)
		//
		for(i=0;i<numport;i++) {			
			if (efd[i] != -1 && FD_ISSET( efd[i], &rfds )) {
				len = read( efd[i], ptr[i], rlen[i] );

				if (len <= 0) {
					printf("efd[%d] read error %d\n", i, len);
					close( efd[i] );
					iidx[i] = 0;
					efd[i] = connect_serv( &ip, rs.port[i] , &cfd[i] );
					continue;
				}

				iopp = (iopacket *)&bigbuf[i][0];
				
				while( ispacket( iopp, ptr[i] + len ) ) {

					if ( (iopp->magic & MAGIC_MASK) != MAGIC) {
						printf("### port %d magic error ! close and re-connect\n",i);
						close(efd[i]);
						
						ptr[i] = &bigbuf[i][0];
						rlen[i] = BIGBUFLEN;
						
						efd[i] = connect_serv( &ip, rs.port[i], &cfd[i] );
						goto	NEXTPORT;
					}

					if (iopp->magic == CMD_IOCTL) {
						if (iidx[i] == iopp->len) {
						
							ioctl( pfd[i], IOC_RET, iopp );
							iopp = (iopacket*)((char *)iopp + HEADER + IOCTL_BODY);
							send_cmd( efd[i], CMD_IOCTLACK, genbuf);
							iidx[i]++;
						} else
							printf("### wrong idx %d\n", iopp->len);
					} else
					if (iopp->magic == CMD_DATA) {
						write(pfd[i], iopp->val.ptr, iopp->len);
						n = iopp->len + 3 + HEADER;
						n &= ~3;
						
						replyflag[i] = 1;
						iopp = (iopacket*)( (char *)iopp + n );
					} else
					if (iopp->magic == CMD_ACK) {
//printf("## ptylock[%d] == 0\n", i);
						ptylock[i] = 0;
						locktime[i] = 0;
						iopp = (iopacket*)((char*)iopp + HEADER);
					} else
					if ((iopp->magic == CMD_CONFIRM) && !replyflag[i]) {
						send_cmd( efd[i], CMD_ACK, genbuf);
//printf("## CONFIRM ACK\n");
						iopp = (iopacket*)((char*)iopp + HEADER);
					} else 
					if (iopp->magic == CMD_IOCTLACK) {
						ilock[i] = 0;
						itime[i] = 0;
						iopp = (iopacket*)((char*)iopp + HEADER);
					}

				}
				
				n = (u32)ptr[i] + len - (u32)iopp;
				ptr[i] = &bigbuf[i][0];
				rlen[i] = BIGBUFLEN;
				
				if (n > ((u32)&bigbuf[i+1][0] - (u32)iopp)) {
					printf("### Serious error !! n = %d\n", (int)n);
					printf("    DEBUG: ptr[%d] %x len %d iopp %x\n",
						i, (unsigned int)ptr[i], len, (unsigned int)iopp);
					exit(1);
				} 
				
				if (n) {
					memcpy( &bigbuf[i][0], iopp, n);
					ptr[i] += n;
					rlen[i] -= n;
				} 
			}
			NEXTPORT:
		}	
		//
		//	datum come from pty
		//
		for(i=0;i<numport;i++) {

			//
			//	just data
			//

			if (FD_ISSET( pfd[i], &rfds )) {

				iopp = (iopacket *) genbuf;
				len = read( pfd[i], iopp->val.ptr , BUFLEN);
				if (len <=0 ) {
					printf("pty port %d read error\n", i);
					close( pfd[i] );
				}
				iopp->cmd = -1;
				iopp->len = len;
				n = len + HEADER + 3;
				n &= ~3;
				iopp->magic = CMD_DATA;
				ret = write( efd[i], iopp , n );
//printf("## ptylock[%d] == 1\n", i);
				ptylock[i] = 1;
				time( &locktime[i] );
				
				if (ret != n) {
					printf("### port %d data not completed write %d %d\n", i, (int)n , ret);
				}
				if (ret == -1) {
					printf("Ethernet port %d write error\n", i);
					close( efd[i] );
					efd[i] = -1;
				}
			}
			
			//
			//	Ioctl from pty
			//
			if (FD_ISSET( pfd[i], &efds )) {

				iopp = &iocbuf[i];

				ioctl( pfd[i], IOC_GET, iopp );

				iopp->magic = CMD_IOCTL;
				iopp->len = iidx[i];
				ilock[i] = 1;
				time( &itime[i] );
				ret = write( efd[i], iopp, IOCTL_BODY + HEADER );
				if (ret != IOCTL_BODY + HEADER) {
					printf("### port %d IOCTL not conpleted write %d %d\n", i , IOCTL_BODY+ HEADER, ret);
				}
				
			}

		}
	} // for(;;)
	return 0;
}
