#define	LOG
#define	MONITOR

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <stdio.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include "rserial.h"


#ifdef	MONITOR

int	msfd[ MAXPORT ];
int	mefd[ MAXPORT ];
int	mlsfd[ MAXPORT ];
int	mlefd[ MAXPORT ];

#endif

#ifdef	LOG

int	e2s[ MAXPORT ];
int	s2e[ MAXPORT ];

#endif

int		sfd[ MAXPORT ];		// Serial port fd
int		lfd[ MAXPORT ];		// Listen fd
int		efd[ MAXPORT ];		// Ethernet fd
int		replyflag[ MAXPORT ];
int		ttylock[ MAXPORT ];
time_t		locktime[ MAXPORT ];

int		iidx[ MAXPORT ];	// ioctl index
int		ilock[ MAXPORT ];	// ioctl lock
time_t		itime[ MAXPORT ];	// ioctl lock time
iopacket	iocbuf[ MAXPORT ];


char		bigbuf[ MAXPORT ][ BIGBUFLEN ];

char		*ptr[ MAXPORT ];
int		rlen[ MAXPORT ];

char		genbuf[ BUFLEN + HEADER ];

int	ispacket( iopacket *iopp, char *end )
{
u32	bound;

	if ((u32)iopp >= (u32)end)
		return 0;


	bound = (u32)iopp->val.ptr + iopp->len;
	
	bound += 3;
	bound &= ~3;
	
	if (bound <= (u32)end)
		return 1;
	else
		return 0;
}

int	do_ioctl( int sfd, int efd, iopacket *iopp)
{
int	ret;

	switch( iopp->cmd ) {
		//
		//	VALUE
		//
		case	TCSBRK:
		case	TCXONC:
		case	TCFLSH:
		case	TIOCEXCL:
		case	TIOCNXCL:
		case	TIOCSCTTY:
		case	TIOCGPGRP:
		case	TIOCSPGRP:
		case	TIOCCONS:
		case	TIOCNOTTY:
		case	TCSBRKP:
		case	TIOCSBRK:
		case	TIOCCBRK:
		case	FIONCLEX:
		case	FIOCLEX:
		case	FIOASYNC:
		case	TIOCSERCONFIG:
		case	TIOCSERGWILD:
		case	TIOCSERSWILD:
		case	TIOCMIWAIT:
			
			ret = ioctl( sfd , iopp->cmd, iopp->val.arg );
			break;
		default:
					
			ret = ioctl( sfd , iopp->cmd, iopp->val.ptr );
	} // switch

#ifdef	DEBUG
	printf("IOCTL: sfd %d cmd %d ret %d\n", sfd, iopp->cmd, ret);
#endif

	iopp->retval = ret;
	iopp->magic = CMD_IOCTL;
	ret = write( efd , iopp , IOCTL_BODY + HEADER );

	if (ret != IOCTL_BODY + HEADER) {
		printf("### IOCTL write not completed %d %d\n", IOCTL_BODY+HEADER, ret);
	}
	
	return 0;
}


int	open_rserial( int fd[], int start )
{
int	num = 0;
char	buf[64];
int	i;

#ifdef	DEBUG

	for(i=0;i<2;i++) {
		sprintf(buf, "/dev/ttyS%d", i );
		fd[i] = open(buf, O_RDWR);
		if (fd[i] < 0)
			break;
		else
			num++;
	}



#else	// DEBUG

	for(i=0;i<MAXPORT;i++) {
		sprintf(buf, "/dev/ttyS%d", i + start);
		fd[i] = open(buf, O_RDWR);
		if (fd[i] < 0)
			break;
		else
			num++;

#ifdef	LOG

		sprintf(buf, "/tmp/s2e.%d", i + start );

		s2e[i] = open(buf, O_CREAT | O_TRUNC | O_WRONLY, 0644);
		
		if (s2e[i] < 0) {
			printf("Unable to Open log file %s\n", buf);
			s2e[i] = -1;
		}
				

#endif	// LOG

	}


#endif	// DEBUG
	
	return num;
}


void	bind_port( int numport, int fd[], int port )
{
int 	i;
int	one = 1;
struct	sockaddr_in	in;

	for(i=0;i < numport; i++) {
		fd[i] = socket(AF_INET, SOCK_STREAM, 0);
		if (fd[i] < 0) {
			printf("Cannot open socket %d\n", i);
			exit(1);
		}

		if ( setsockopt( fd[i] , SOL_SOCKET, SO_REUSEADDR, 
				(char *)&one, sizeof(one)) == -1) {
			printf(" port %d SO_REUSEADDR error\n", i);
			exit(1);
		}

		memset( &in, 0, sizeof(struct sockaddr_in));

		in.sin_family = AF_INET;
		in.sin_addr.s_addr = htonl(INADDR_ANY);	
		in.sin_port = htons( port + i);
		
		if (bind( fd[i], (struct sockaddr *)&in, sizeof(struct sockaddr_in)) < 0) {
			printf("Cannot bind fd %d\n", i);
			exit(1);
		}
		printf("Bind port %d\n", htons(in.sin_port) );
		listen( fd[i], 5);
	}

}

void	send_cmd(int fd, int cmd, void *buf)
{
	((iopacket *)buf)->cmd = cmd;
	((iopacket *)buf)->magic = cmd;
	((iopacket *)buf)->len = 0;
	write( fd, buf, HEADER);
}


int main(int argc, char *argv[])
{
int	serversock;
int	one = 1;
struct	sockaddr_in	in;
struct	sockaddr_in	new;
struct	Rserial	rs;
int	i;
fd_set	rfd, wfd;
int	maxfd;
int	ret;
int	newsock;
int	addr_len;
int	len;
u32	n;
iopacket	*iopp;
int	numport;
struct	timeval	tv;
char	tmpbuf[ HEADER ];

	if (argc > 1) {
		ret = atoi( argv[1] );
		if (ret < 0) {
			printf("illegal port number %d\n", ret);
			exit(1);
		}

	} else
		ret = 44;	

	printf("Multi-serial port devices start from /dev/ttyS%d\n", ret);

	serversock = socket(AF_INET, SOCK_STREAM, 0);
	if (serversock < 0) {
		printf("Cannot open socket \n");
		exit(1);
	}

	if ( setsockopt( serversock, SOL_SOCKET, SO_REUSEADDR, 
			(char *)&one, sizeof(one)) == -1) {
		printf("setsockopt SO_REUSEADDR error\n");
		exit(1);
	}

	memset( &in, 0 , sizeof(struct sockaddr_in) );

	in.sin_family = AF_INET;
	in.sin_addr.s_addr = htonl(INADDR_ANY);
	in.sin_port = htons(SERVER_PORT);
	
	if (bind( serversock, (struct sockaddr *)&in, sizeof(struct sockaddr_in))<0) {
		printf("Cannot bind server port \n");
		exit(1);
	}
		
	listen( serversock, 5);

	printf("listen\n");


	rs.num = numport = open_rserial( sfd, ret );

	if ( !numport ) {
		printf("No multi-serial port found.\n");
		exit(1);
	} else 
		printf("Port number : %d\n", numport);

	for(i=0;i<MAXPORT;i++) {
		rs.port[i] = SERVER_PORT + i + 1;
		efd[i] = -1;

#ifdef	MONITOR
		msfd[i] = -1;
		mefd[i] = -1;
		mlsfd[i] = -1;
		mlefd[i] = -1;
#endif
	}

	bind_port( numport , lfd , SERVER_PORT + 1 );

#ifdef	MONITOR

	bind_port( numport, mlsfd, MONITOR_PORT );
	bind_port( numport, mlefd, MONITOR_PORT + 100 );
#endif

	addr_len = sizeof(struct sockaddr_in);

	for(i=0;i<numport;i++) {
		ptr[i] = &bigbuf[i][0];
		rlen[i] = BIGBUFLEN;
	}
	memset( ttylock, 0, sizeof(ttylock));
	memset( replyflag, 0, sizeof(replyflag));
	memset( locktime, 0, sizeof(locktime));
	memset( iidx, 0, sizeof(iidx));
	memset( ilock, 0, sizeof(ilock));
	memset( itime, 0, sizeof(itime));		




	for(;;) {
		
		FD_ZERO( &rfd );
		FD_ZERO( &wfd );
		FD_SET( serversock, &rfd);

		maxfd = serversock;
		
		//
		//	FD_SET	Serial fd
		//
		for(i=0;i<numport;i++) {
			if (sfd[i] != -1) {
				if ( !ttylock[i] )
					FD_SET( sfd[i], &rfd);
					
				FD_SET( sfd[i], &wfd);
				maxfd = (maxfd > sfd[i]) ? maxfd : sfd[i];
			}

		}
		
		//
		//	FD_SET	listen fd
		//	Listen to all fds. Maybe client loses connect and 
		//	make another connection.
		//
		for(i=0;i<numport;i++) {
			FD_SET( lfd[i], &rfd);
			maxfd = (maxfd > lfd[i]) ? maxfd : lfd[i];
		}
		
		//
		//	FD_SET	Ethernet fd
		//
		for(i=0;i<numport;i++) {
			if (efd[i] != -1) {
				FD_SET( efd[i], &rfd);
				maxfd = (maxfd > efd[i]) ? maxfd : efd[i];
			}
		}

#ifdef	MONITOR
		//
		//	FD_SET	Monitor, the listen port
		//
		for(i=0;i<numport;i++) {
			FD_SET( mlsfd[i], &rfd);
			FD_SET( mlefd[i], &rfd);
			maxfd = (maxfd > mlsfd[i]) ? maxfd : mlsfd[i];
			maxfd = (maxfd > mlefd[i]) ? maxfd : mlefd[i];	
		}
#endif		

		tv.tv_sec = 1;
		tv.tv_usec = 0;
		
		ret = select(maxfd + 1, &rfd, &wfd, NULL, &tv);

		if (ret < 0) {
			printf("Select error !!\n");
			exit(1);
		}

		//
		//	CONFIRM
		//
		for(i=0;i<numport;i++) {
			if (locktime[i] && (( time(NULL)-locktime[i] )>CONFIRM_T) ) {
printf("# CMD_CONFIRM\n");
				send_cmd( efd[i], CMD_CONFIRM, tmpbuf);
				time( &locktime[i] );
			}

			if (itime[i] && (( time(NULL)-itime[i] ) > CONFIRM_T ) ) {

				// CMD_IOCTLACK not coming, resend ioctl cmd again
				
				write( efd[i], &iocbuf[i], IOCTL_BODY + HEADER );
				time( &itime[i] );
			}
				

		}

		if (ret == 0) {

			continue;	
		}

#ifdef	MONITOR
		//
		//	Here comes the Monitor
		//
		for(i=0;i<numport;i++) {
			if (FD_ISSET( mlsfd[i], &rfd)) {
				if (msfd[i] != -1)
					close(msfd[i]);
				newsock = accept( mlsfd[i], (struct sockaddr *)&new, (int*)&addr_len);
				if (newsock >= 0) {
					msfd[i] = newsock;
					printf("Accept new Monitor for Serial port %d\n", i);
				}
			}

			if (FD_ISSET( mlefd[i], &rfd)) {
				if (mefd[i] != -1)
					close(mefd[i]);
				newsock = accept( mlefd[i], (struct sockaddr *)&new, (int*)&addr_len);
				if (newsock >= 0) {
					mefd[i] = newsock;
					printf("Accept new Monitor for Ethernet %d\n", i);
				}
			}

		}

#endif
		//
		//	Reply with ACK
		//
		for(i=0;i<numport;i++) {
			if ( replyflag[i] && FD_ISSET( sfd[i], &wfd ) ) {
				send_cmd( efd[i], CMD_ACK, tmpbuf);
				replyflag[i] = 0;
			}
		}


		//
		//	Query Server info.
		//
		
		if (FD_ISSET( serversock, &rfd)) {
			newsock = accept(serversock, (struct sockaddr *)&new, (int*)&addr_len);
			
			if (newsock >= 0) {
				printf("Query server info from %s\n", inet_ntoa(new.sin_addr));
				write( newsock, &rs, sizeof(struct Rserial));
				close(newsock);
			}
		}
			
		//
		//	Accept new connection
		//	
		for(i=0;i<numport;i++) {
			if ( FD_ISSET( lfd[i], &rfd)) {
				if (efd[i] != -1)	// client lost connecton
					close(efd[i]);
				newsock = accept( lfd[i], (struct sockaddr *)&new, (int*)&addr_len);

				if (newsock < 0)
					continue;

#ifdef	LOG
				sprintf(genbuf, "/tmp/e2s.%d", i);
				e2s[i] = open(genbuf, O_CREAT | O_TRUNC | O_WRONLY, 0644);

				if ( e2s[i] < 0) {
					printf("Unable to open log file :%s\n", genbuf);
					e2s[i] = -1;
				}

#endif

				efd[i] = newsock;
				printf("Accept new connection %d\n", i);
				iidx[i] = 0;
				ret = setsockopt( newsock, SOL_SOCKET, SO_KEEPALIVE, &one, sizeof one);
				if (ret)
					printf("setsockopt SO_KEEPALIVE error %d\n", i);

			}
		}


		//
		//	Datum come from serial port
		//			
		for(i=0;i<numport;i++) {
			if (sfd[i] != -1 && FD_ISSET( sfd[i], &rfd)) {
				iopp = (iopacket *) genbuf ;
				len = read( sfd[i], iopp->val.ptr , BUFLEN);
				if (len < 0) {
					printf("Serial port %d read error\n", i);
					continue;
				}
				
				if (len == 0) {
					printf("Serial port %d read 0, EOF\n", i);
					continue;
				}
#ifdef	MONITOR
				if (msfd[i] != -1) {
					ret = write(msfd[i], iopp->val.ptr, len);
					if (ret <= 0) {
						close(msfd[i]);
						msfd[i] = -1;
					}
				}
#endif
#ifdef	LOG
				if (s2e[i] != -1)
					write(s2e[i], iopp->val.ptr, len);
#endif

				if (efd[i] == -1) {
					printf("efd[%d] is -1, unable to write datum to ethernet.\n", i);
					continue;
				} else {
					iopp->cmd = -1;
					iopp->len = len;
					n = len + HEADER + 3;
					n &= ~3;
					iopp->magic = CMD_DATA;

					ret = write(efd[i], iopp , n );
					ttylock[i] = 1;
					time( &locktime[i] );

					if (ret != n) {
						printf("### port %d write to ethernet not completed %d %d\n", 
							i, (int)n, ret);
					}
					
					if (ret <= 0) {
						close( efd[i] );
						efd[i] = -1;
					}
				}
			
			}
		}

		
		//
		//	Datum come from Ethernet
		//
		//	1. Data	 (cmd == -1)
		//	2. Ioctl (cmd != -1)
		//	3. ACK.  (cmd == -1) && (len == 0)
		//
		for(i=0;i<numport;i++) {
			if ( efd[i] != -1 && FD_ISSET( efd[i], &rfd)) {
				len = read( efd[i], ptr[i], rlen[i] );

				if (len < 0) {
					printf("port %d read error \n", i);
					close( efd[i] );
					efd[i] = -1;
					continue;
				}
				if (len == 0) {
					printf("port %d closed or lose connection\n", i );
					close( efd[i] );
					efd[i] = -1;
				}
					
				iopp = (iopacket *)&bigbuf[i][0];
				
				while( ispacket( iopp, ptr[i] + len ) ) {

					if ( (iopp->magic & MAGIC_MASK) != MAGIC) {
						printf("### port %d magic error ! close client to let she re-connect\n", i);
						close(efd[i]);
						
						ptr[i] = &bigbuf[i][0];
						rlen[i] = BIGBUFLEN;
						
						efd[i] = -1;
						goto	NEXTPORT;
					}

					if (iopp->magic == CMD_DATA) {
						if ( !FD_ISSET( sfd[i], &wfd) ) {
							printf("###### Serious ERROR: sfd[%d] not ready for write !\n", i);
						}

#ifdef	MONITOR
						if ((mefd[i] != -1) && (iopp->len)) {
							ret = write(mefd[i], iopp->val.ptr, iopp->len);
							if (ret <= 0) {
								close( mefd[i] );
								mefd[i] = -1;
							}
						}
	
#endif
#ifdef	LOG
						if (e2s[i] != -1)
							write(e2s[i], iopp->val.ptr, iopp->len);
						
#endif
						ret = write(sfd[i], iopp->val.ptr, iopp->len);
						
						if (ret == 0) {
							printf("##### ERROR: write return 0 sfd[%d]\n", i);
						} else 
						if (ret < 0) {
							printf("##### ERROR: write error %d sfd[%d]\n", ret, i);
						} else
						if (ret < iopp->len) {
							printf("#### WARNNING: write not completed sfd[%d] %d %d\n", i, ret, iopp->len);
						}	
						
						n = iopp->len + 3 + HEADER;
						n &= ~3;
						
						replyflag[i] = 1;
						iopp = (iopacket*)( (char *)iopp + n );
					} else
					if (iopp->magic == CMD_IOCTL) {
						if (iidx[i] == iopp->len) {
							memcpy( &iocbuf[i] , iopp, sizeof(iopacket)); // save or clone
						
							do_ioctl( sfd[i], efd[i], &iocbuf[i] );
						
						
							ilock[i] = 1;
							time( &itime[i] );
						
							iopp = (iopacket*)((char *)iopp + HEADER + IOCTL_BODY);
						
							send_cmd( efd[i], CMD_IOCTLACK, tmpbuf);
						
							iidx[i]++;
						} else 
							printf("### Idx error %d\n", iopp->len );
					} else
					if (iopp->magic == CMD_ACK) {
//printf("## ttylock[%d] == 0\n", i);
						ttylock[i] = 0;
						locktime[i] = 0;
						iopp = (iopacket*)((char*)iopp + HEADER);
					} else
					if ((iopp->magic == CMD_CONFIRM) && !replyflag[i]) {
//printf("## CONFIRM ACK\n");
						send_cmd( efd[i], CMD_ACK, tmpbuf);
						iopp = (iopacket*)((char*)iopp + HEADER);
					} else
					if (iopp->magic == CMD_IOCTLACK) {
						ilock[i] = 0;
						itime[i] = 0;
						iopp = (iopacket*)((char*)iopp + HEADER);
					}
				}
				
				n = (u32)ptr[i] + len - (u32)iopp;
				ptr[i] = &bigbuf[i][0];
				rlen[i] = BIGBUFLEN;
				
				if (n) {
					memcpy( &bigbuf[i][0], iopp, n);
					ptr[i] += n;
					rlen[i] -= n;
				} 
			}
			NEXTPORT:
		}
	} // for(;;)
}

