
#define	VERBOSE

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/init.h> 
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/major.h>
#include <linux/param.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <linux/ioctl.h>
#include <linux/serial.h>
#include <linux/serialP.h>
#include <linux/tty.h>

#if LINUX_VERSION_CODE > 0x020400
#include <linux/devfs_fs_kernel.h>
#endif

#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#define	DEBUG


#define	MAJOR_TTYRS	253
#define	MAJOR_PTYRS	254
#define	NUMPORT		16

static	char *ttyrs	= "ttyrs/%d";
static	char *ptyrs	= "ptyrs/%d";

#define	RS_IOC_MAGIC	'Z'
#define	RS_IOCGET	_IOR( RS_IOC_MAGIC, 1, iopacket )
#define	RS_IOCRET	_IOW( RS_IOC_MAGIC, 1, iopacket )

//
//	This the ioctl command packet transfing between local and remote
//
#include "rserial.h"

//
//	This is ioctl command set used inside kernel
//

typedef	struct	{
	iopacket		iopack;
	int			flag;
	int			ret;
#if LINUX_VERSION_CODE > 0x020400
	wait_queue_head_t	iowait;		// ioctl wait queue
#else
	struct	wait_queue	*iowait;
#endif
} ioctl_set ;

ioctl_set	Iocset[NUMPORT];

/*
	Buffer Usage:

		Write to its buffer and read from others buffer.
 */

struct 	rs_private	{

#if LINUX_VERSION_CODE > 0x020400
	wait_queue_head_t	rwait, wwait;	// read and write wait queue
#else
	struct	wait_queue	*rwait, *wwait;
#endif
	char			*buffer, *end;
	char			*rp, *wp;	// local read and write pointer
	int			full;		// If the buffer is full
	int			flag;		// if the device is opened or not
	ioctl_set		*iocset;	// pty and tty has the same copy
	
	struct	fasync_struct	*async_queue;
	struct rs_private	*rdev;		// remote device
};


struct	rs_private	tty_private[NUMPORT];
struct	rs_private	pty_private[NUMPORT];

#define	BUFSIZE	10240

static	int	MIN(int x,int y)
{
	if (x > y)
		return y;
	else
		return x;
}

static int tty_open(struct inode * inode, struct file * filp)
{
int minor = MINOR(inode->i_rdev);
struct rs_private *dev;

	if ((minor > NUMPORT) || (minor < 0))
		return	-ENODEV;

	if ( !pty_private[minor].flag )
		return	-ENODEV;

	dev = &tty_private[minor];

	if (!filp->private_data)
		filp->private_data = (void *) dev;

	//
	//	First Open, reset buffer of tty
	
	if ( !dev->flag ) {
		dev->full = 0;
		dev->rp = dev->wp = dev->buffer;
	}

	dev->flag++;	
	
	wake_up_interruptible( &dev->rwait );

#ifndef	DEBUG
	MOD_INC_USE_COUNT;
#endif

	return 0;
}

static int pty_open(struct inode * inode, struct file * filp)
{
int minor = MINOR(inode->i_rdev);
struct rs_private *dev, *rdev;	

	if ( (minor > NUMPORT) || (minor < 0) )
		return	-ENODEV;

	dev	= &pty_private[minor];
	rdev	= &tty_private[minor];

	dev->rdev = rdev;
	rdev->rdev = dev;

	dev->iocset = rdev->iocset = &Iocset[ minor ];
	
	if (!dev->buffer) {
		dev->buffer = kmalloc( BUFSIZE, GFP_KERNEL);
		
		if (!dev->buffer)
			return	-EFAULT;
		
#if LINUX_VERSION_CODE > 0x020400
		init_waitqueue_head( &dev->rwait );
		init_waitqueue_head( &dev->wwait );	
#endif
		dev->end = dev->buffer + BUFSIZE;

		dev->rp = dev->wp = dev->buffer;
		dev->full = 0;
		
		rdev->buffer = kmalloc( BUFSIZE, GFP_KERNEL);

#if LINUX_VERSION_CODE > 0x020400
		init_waitqueue_head( &rdev->rwait );
		init_waitqueue_head( &rdev->wwait );
#endif
		rdev->end = rdev->buffer + BUFSIZE;

		rdev->rp = rdev->wp = rdev->buffer;			
		rdev->full = 0;
	}	
	
	if (!filp->private_data)
		filp->private_data = (void *) dev;

	dev->flag++;

#ifndef	DEBUG
	MOD_INC_USE_COUNT;
#endif
	return 0;
}

//
// The following release routines can be merged into one function
// in the future. 			William Chen


static	int tty_release(struct inode * inode, struct file * filp)
{
struct rs_private *dev = filp->private_data;
		
	wake_up_interruptible( &dev->rwait );
	wake_up_interruptible( &dev->rdev->wwait );

	dev->flag--;
	if (dev->flag < 0)
		dev->flag = 0;

#ifndef	DEBUG
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}

static	int pty_release(struct inode * inode, struct file * filp)
{
struct rs_private *dev = filp->private_data;

	wake_up_interruptible( &dev->rwait );
	wake_up_interruptible( &dev->rdev->wwait );

	dev->flag--;
	if (dev->flag < 0)
		dev->flag = 0;
		
	if (!dev->flag) {
		dev->full = 0;
		dev->rp = dev->wp = dev->buffer;
	}

#ifndef	DEBUG
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}
	//
	// Read the datum in pty's buffer
	//

static ssize_t	tty_read(struct file *filp, char *buf,
			 size_t count, loff_t *off )
{
struct rs_private *dev = filp->private_data;
int	n = 0;

	dev = dev->rdev;

	if (!dev->flag)
		return -ENODEV;

	while( (dev->rp == dev->wp) && !dev->full ){	// Empty
		if (filp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		if (wait_event_interruptible( dev->rwait, (dev->rp != dev->wp) || dev->full))
			return -ERESTARTSYS;
	}

	if (!dev->flag)
		return -ENODEV;

	if (dev->wp > dev->rp)
		count = MIN(count, dev->wp - dev->rp);
	else {
 		n = count;
 		count = MIN(count, dev->end - dev->rp);
		n -= count;
		if (n > 0)
			n = MIN(n, dev->wp - dev->buffer);
		else
			n = 0;
	}
		
	if (copy_to_user( buf, dev->rp, count))
		return -EFAULT;
	dev->rp += count;

	if (dev->rp == dev->end)
		dev->rp = dev->buffer;

	if (n && copy_to_user(buf+count, dev->buffer, n))
		return -EFAULT;
	dev->rp += n;
		
	// redundant
	
	if (dev->rp == dev->end)
		dev->rp = dev->buffer;

	dev->full = 0;
	
	wake_up_interruptible( &dev->wwait );
	
	return count + n;
}

static ssize_t pty_read(struct file *filp, char *buf, 
			size_t count, loff_t *off )
{
struct rs_private *dev = filp->private_data;
int	n = 0;

	dev = dev->rdev;

	while( (dev->rp == dev->wp) && !dev->full ) {	// Empty
		if (filp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		if (wait_event_interruptible( dev->rwait, (dev->rp != dev->wp) || dev->full))
			return -ERESTARTSYS;
	}

	if (dev->wp > dev->rp)
		count = MIN(count, dev->wp - dev->rp);
	else {
		n = count;
		count = MIN(count, dev->end - dev->rp);
		n -= count;
		if (n > 0) 
			n = MIN(n, dev->wp - dev->buffer);
		else
			n = 0;
	}
	
	if (copy_to_user( buf, dev->rp, count))
		return -EFAULT;
	dev->rp += count;

	if (dev->rp == dev->end)
		dev->rp = dev->buffer;

	if (n && copy_to_user(buf+count, dev->buffer, n))
		return -EFAULT;
	dev->rp += n;
		
	// redundant
	
	if (dev->rp == dev->end)
		dev->rp = dev->buffer;

	dev->full = 0;
	
	wake_up_interruptible( &dev->wwait );
	
	return count + n;
}

static ssize_t tty_write(struct file * filp, const char * buf,
		        size_t count, loff_t *off)
{
struct rs_private *dev = filp->private_data;
int	n = 0;

	if (!dev->rdev->flag)
		return -ENODEV;

	while( dev->full ) {
		if (filp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		if (wait_event_interruptible( dev->wwait, !dev->full ) )
			return -ERESTARTSYS;
	}

	if (!dev->rdev->flag)	// after waiting, pty is closed
		return -ENODEV;

	if (dev->rp > dev->wp) {
		count = MIN(count, dev->rp - dev->wp);
	} else {
		n = count;
		count = MIN(count, dev->end - dev->wp);
		n -= count;
		if (n > 0)
			n = MIN(n, dev->rp - dev->buffer);
		else
			n = 0;
	}

	if (copy_from_user( dev->wp, buf, count ) ) 
		return -EFAULT;
		
	dev->wp += count;
	if (dev->wp == dev->end)
		dev->wp = dev->buffer;
	
	if (n && copy_from_user( dev->wp, buf+count, n))
		return -EFAULT;
	dev->wp += n;
	
	// redundant
	if (dev->wp == dev->end)
		dev->wp = dev->buffer;
		
	if (dev->rp == dev->wp)
		dev->full = 1;
		
	wake_up_interruptible( &dev->rwait );
	
	if (dev->async_queue)
#if LINUX_VERSION_CODE > 0x020400
		kill_fasync( &dev->async_queue, SIGIO, POLL_IN);
#else
		kill_fasync( dev->async_queue, SIGIO, POLL_IN);
#endif
	return count + n;
}

static ssize_t pty_write(struct file * filp, const char * buf,
		        size_t count, loff_t *off)
{
struct rs_private *dev = filp->private_data;
struct rs_private *rdev = dev->rdev;
int	n = 0;

	while( dev->full && rdev->flag ) {
		if (filp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		if (wait_event_interruptible( dev->wwait, !dev->full || !rdev->flag ))
			return -ERESTARTSYS;
	}

	if (!rdev->flag) {
		dev->rp = dev->wp = dev->buffer;
		dev->full = 0;
		return count;
	}

	if (dev->rp > dev->wp) {
		count = MIN(count, dev->rp - dev->wp);
	} else {
		n = count;
		count = MIN(count, dev->end - dev->wp);
		n -= count;
		if (n > 0)
			n = MIN(n, dev->rp - dev->buffer);
		else
			n = 0;
	}

	if (copy_from_user( dev->wp, buf, count ) ) 
		return -EFAULT;
		
	dev->wp += count;
	if (dev->wp == dev->end)
		dev->wp = dev->buffer;
	
	if (n && copy_from_user( dev->wp, buf+count, n))
		return -EFAULT;
	dev->wp += n;
	
	// redundant
	if (dev->wp == dev->end)
		dev->wp = dev->buffer;
		
	if (dev->rp == dev->wp)
		dev->full = 1;
	
	wake_up_interruptible( &dev->rwait );
	
	if (dev->async_queue)
#if LINUX_VERSION_CODE > 0x020400
		kill_fasync( &dev->async_queue, SIGIO, POLL_IN);
#else
		kill_fasync( dev->async_queue, SIGIO, POLL_IN);
#endif
	return count + n;

}

static unsigned int tty_poll(struct file *filp, poll_table *wait)
{
struct rs_private *dev = filp->private_data;
struct rs_private *rdev = dev->rdev;
unsigned int mask = 0;

	poll_wait(filp, &dev->wwait, wait);
	poll_wait(filp, &rdev->rwait, wait);
	
	if (!rdev->flag)	// pty closed
		return POLLERR;

	//
	//	Read: check remote
	//
	
	if ( (rdev->rp != rdev->wp) || rdev->full )
		mask |= POLLIN | POLLRDNORM;
	
	//
	//	Write: check local
	//
	if ( !dev->full )
		mask |= POLLOUT | POLLWRNORM;
		
	return mask;	
}


static unsigned int pty_poll(struct file *filp, poll_table *wait)
{
struct rs_private *dev = filp->private_data;
struct rs_private *rdev = dev->rdev;
ioctl_set	*iop = dev->iocset;
unsigned int mask = 0;

	poll_wait(filp, &dev->wwait, wait);
	poll_wait(filp, &rdev->rwait, wait);

	//
	//	Read: check remote
	//
	
	if ( (rdev->rp != rdev->wp) || rdev->full )
		mask |= POLLIN | POLLRDNORM;

	//
	//	Write: check local
	//

	if ( !dev->full )
		mask |= POLLOUT | POLLWRNORM;

	if (iop->flag) {
		iop->flag = 0;
		mask |= POLLPRI;
	}
	
	return mask;
}

static int tty_ioctl(struct inode *inode, struct file *filp,
		unsigned int cmd, unsigned long arg)
{
struct	rs_private *dev = filp->private_data;
ioctl_set	*iop = dev->iocset;
iopacket	*iopp = &iop->iopack;

	iopp->cmd = cmd;

	switch( cmd )
	{	
		//
		//	Read from Kernel
		//
		case	TCGETS:
		case	TCGETA:
		case	TIOCOUTQ:
		case	TIOCGWINSZ:
		case	TIOCMGET:
		case	TIOCGSOFTCAR:
		case	TIOCINQ:
		case	TIOCGSERIAL:
		case	TIOCGETD:
		case	TIOCTTYGSTRUCT:
		case	TIOCGSID:
		case	TIOCGPTN:
//		case	TIOCGDEV:
		case	TIOCGLCKTRMIOS:
		case	TIOCSERGSTRUCT:
		case	TIOCSERGETLSR:
		case	TIOCSERGETMULTI:
		case	TIOCGICOUNT:

			iopp->val.arg = arg;

			break;
			
		//
		//	Wrtie to Kernel
		//
			//
			//	struct	termios
			//
		case	TCSETS:
		case	TCSETSW:
		case	TCSETSF:
		case	TCSETA:
		case	TCSETAW:
		case	TCSETAF:
		case	TIOCSLCKTRMIOS:
			if (copy_from_user( &iopp->val.arg, (char*)arg, sizeof(struct termios)))
				return -EFAULT;
			break;
			
			//
			//	int
			//
		case	TIOCSTI:
		case	TIOCMBIS:
		case	TIOCMBIC:
		case	TIOCMSET:
		case	TIOCSSOFTCAR:
		case	TIOCLINUX:
		case	TIOCPKT:
		case	FIONBIO:
		case	TIOCSETD:
		case	TIOCSPTLCK:
			if (copy_from_user( &iopp->val.arg, (char*)arg, sizeof(int)))
				return -EFAULT;
			break;
			
			//
			//	struct	winsize
			//
		case	TIOCSWINSZ:
			if (copy_from_user( &iopp->val.arg, (char*)arg, sizeof(struct winsize)))
				return -EFAULT;
			break;
			
			//
			//	struct	serial_struct
			//
		case	TIOCSSERIAL:
			if (copy_from_user( &iopp->val.arg, (char*)arg, sizeof(struct serial_struct)))
				return -EFAULT;
			break;
			
			//
			//	struct	serial_miltiport_struct
			//
#ifdef	CONFIG_SERIAL_MULTIPORT
		case	TIOCSERSETMULTI:
			if (copy_from_user( &iopp->val.arg, (char*)arg, sizeof(struct serial_multiport_struct)))
				return -EFAULT;	
			break;
#endif

		//
		//	VALUE
		//
		
		case	TCSBRK:
		case	TCXONC:
		case	TCFLSH:
		case	TIOCEXCL:
		case	TIOCNXCL:
		case	TIOCSCTTY:
		case	TIOCGPGRP:
		case	TIOCSPGRP:
		case	TIOCCONS:
		case	TIOCNOTTY:
		case	TCSBRKP:
		case	TIOCSBRK:
		case	TIOCCBRK:
		case	FIONCLEX:
		case	FIOCLEX:
		case	FIOASYNC:
		case	TIOCSERCONFIG:
		case	TIOCSERGWILD:
		case	TIOCSERSWILD:
		case	TIOCMIWAIT:
//		case	FIOQSIZE:
			iopp->val.arg = arg;
			
			break;

		default:
			printk("### -ENOIOCTLCMD %x\n", cmd);
			return -ENOIOCTLCMD;
	}

	iop->ret  = 0;
	iop->flag = 1;				// mark ioctl triggled
	wake_up_interruptible( &dev->rwait );	// wake up pty
	wait_event_interruptible( iop->iowait, iop->ret );
	iop->ret = 0;				// clear flag for wait_event..
	if (iopp->retval < 0)			// Error occurs in remote site
		return iopp->retval;
		

	switch( cmd )
	{	
		//
		//	VALUE
		//
		case	TCSBRK:
		case	TCXONC:
		case	TCFLSH:
		case	TIOCEXCL:
		case	TIOCNXCL:
		case	TIOCSCTTY:
		case	TIOCGPGRP:
		case	TIOCSPGRP:
		case	TIOCCONS:
		case	TIOCNOTTY:
		case	TCSBRKP:
		case	TIOCSBRK:
		case	TIOCCBRK:
		case	FIONCLEX:
		case	FIOCLEX:
		case	FIOASYNC:
		case	TIOCSERCONFIG:
		case	TIOCSERGWILD:
		case	TIOCSERSWILD:
		case	TIOCMIWAIT:
//		case	FIOQSIZE:

			break;

		//
		//	WRITE
		//
		case	TCSETS:
		case	TCSETSW:
		case	TCSETSF:
		case	TCSETA:
		case	TCSETAW:
		case	TCSETAF:
		case	TIOCSLCKTRMIOS:

		case	TIOCSTI:
		case	TIOCMBIS:
		case	TIOCMBIC:
		case	TIOCMSET:
		case	TIOCSSOFTCAR:
		case	TIOCLINUX:
		case	TIOCPKT:
		case	FIONBIO:
		case	TIOCSETD:
		case	TIOCSPTLCK:

		case	TIOCSWINSZ:

		case	TIOCSSERIAL:
			break;
			
		//
		//	READ
		//
			//
			//	int
			//
		case	TIOCOUTQ:
		case	TIOCMGET:
		case	TIOCGSOFTCAR:
		case	TIOCINQ:
		case	TIOCGETD:
		case	TIOCGSID:
		case	TIOCGPTN:
//		case	TIOCGDEV:
		case	TIOCSERGETLSR:

			return copy_to_user( (char*)arg, &iopp->val.arg, sizeof(int));

			//
			//	struct	termios
			//
		case	TCGETS:
		case	TCGETA:
		case	TIOCGLCKTRMIOS:
			if (copy_to_user((char*)arg, &iopp->val.arg, sizeof(struct termios)))
				return -EFAULT;
			break;
			
			//
			//	struct	winsize
			//
		case	TIOCGWINSZ:			
			if (copy_to_user((char*)arg, &iopp->val.arg, sizeof(struct winsize)))
				return -EFAULT;
			break;
			
			//
			//	struct	serial_struct
			//
		case	TIOCGSERIAL:
			if (copy_to_user((char*)arg, &iopp->val.arg, sizeof(struct serial_struct)))
				return -EFAULT;
			break;

			//
			//	struct	tty_struct
			//
		case	TIOCTTYGSTRUCT:
			if (copy_to_user((char*)arg, &iopp->val.arg, sizeof(struct tty_struct)))
				return -EFAULT;
			break;

			//
			//	struct	async_struct
			//
		case	TIOCSERGSTRUCT:
			if (copy_to_user((char*)arg, &iopp->val.arg, sizeof(struct async_struct)))
				return -EFAULT;
			break;

			//
			//	struct	serial_multiport_struct
			//
#ifdef	CONFIG_SERIAL_MULTIPORT
		case	TIOCSERGETMULTI:
			if (copy_to_user((char*)arg, &iopp->val.arg, sizeof(struct serial_multiport_struct)))
				return -EFAULT;
			break;
#endif

			//
			//	struct	serial_icounter_struct
			//
		case	TIOCGICOUNT:
			if (copy_to_user((char*)arg, &iopp->val.arg, sizeof(struct serial_icounter_struct)))
				return -EFAULT;
			break;

		
		default:
			printk("### ioctl return -ENOIOCTLCMD %x\n", cmd);
			return -ENOIOCTLCMD;

	}
	return iopp->retval;
}

static int pty_ioctl(struct inode *inode, struct file *filp,
		unsigned int cmd, unsigned long arg)
{
struct	rs_private *dev = filp->private_data;
ioctl_set	*iop = dev->iocset;
iopacket	*iopp = &iop->iopack;

	switch( cmd ) {
		case	0:
		case	RS_IOCGET:
			copy_to_user( (unsigned int *)arg, iopp, 
					sizeof(iopacket) );
			break;
			
		case	1:
		case	RS_IOCRET:
			copy_from_user( iopp, (unsigned int *)arg,
					sizeof(iopacket) );
			iop->ret = 1;
			wake_up_interruptible( &iop->iowait );
	}
	return 0;
}


static struct	file_operations	ttyrs_op = {
	write:		tty_write,
	read:		tty_read,
	open:		tty_open,
	release:	tty_release,
	poll:		tty_poll,
	ioctl:		tty_ioctl,
};


static struct	file_operations	ptyrs_op = {
	write:		pty_write,
	read:		pty_read,
	open:		pty_open,
	release:	pty_release,
	poll:		pty_poll,
	ioctl:		pty_ioctl,
};

#if LINUX_VERSION_CODE > 0x020400
static int rs_init(void)
#else
int	init_module(void)
#endif
{
int	ret = 0;

	memset( tty_private, 0, sizeof(tty_private) );
	memset( pty_private, 0, sizeof(pty_private) );	

	memset( Iocset, 0, sizeof(Iocset) );
#if LINUX_VERSION_CODE > 0x020400
	for(ret=0;ret<NUMPORT; ret++)
		init_waitqueue_head( &Iocset[ret].iowait );
#endif
	
	if ( (ret = register_chrdev( MAJOR_TTYRS, ttyrs, &ttyrs_op))) {
		printk("rserial: register remote serial tty faild..\n");
		return ret;
	}

	if ( (ret = register_chrdev( MAJOR_PTYRS, ptyrs, &ptyrs_op))) {
		printk("rserial: register remote serial pty failed..\n");
		return ret;
	}

	printk("Remote serial port registered.\n");
	return ret;
}

#if LINUX_VERSION_CODE > 0x020400
static void rs_exit(void)
#else
void cleanup_module(void)
#endif
{
int	ret;
int	i;
	
	for(i=0;i<NUMPORT;i++) {
		if (tty_private[i].buffer)
			kfree( tty_private[i].buffer );
			
		if (pty_private[i].buffer)
			kfree( pty_private[i].buffer );
	}

	if ( ( ret = unregister_chrdev( MAJOR_TTYRS, ttyrs))) {
		printk("rserial: unregister remote serial tty failed..\n");
		return;
	}
	
	if ( ( ret = unregister_chrdev( MAJOR_PTYRS, ptyrs))) {
		printk("rserial: unregister remote serial pty failed..\n");
		return;
	}
	printk("Remote serial port unregistered.\n");

}

#if LINUX_VERSION_CODE > 0x020400

module_init( rs_init );
module_exit( rs_exit );

#endif
