

#define	BUFSIZE		4096

#define	TEST_SIZE	128*1024

#define	MAXPORT		8

#define	CHUNK_SIZE	256		// for all ASCII

#define	GLOBLOFFSET	3
#define	PORTROOM	2

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <time.h>
#include <ctype.h>
#include <string.h>

#define	ESC	27
#define	clrscr()	printf("%c[2J", ESC)
#define	clrlin()	printf("%c[K", ESC)
#define	gotoxy(x, y)	printf("%c[%d;%dH", ESC, x, y)
	

#define	LOG_FILE

#ifdef	LOG_FILE

int	logfd[ MAXPORT ];

#endif

int	open_dev( char *name, int num)
{
int	fd;
char	buf[100];


	sprintf(buf, "%s%d", name, num);
	
	fd = open(buf, O_RDWR);
	if (fd < 0) {
		printf("Unable to open %s\n", buf);
		exit(1);
	}

#ifdef	LOG_FILE

	sprintf(buf, "/tmp/rswrite.%d", num);
	logfd[num] = open(buf, O_CREAT | O_TRUNC | O_WRONLY, 0644);

	if (logfd[num] < 0) {
		printf("Unable to open log file :%s\n", buf);
		logfd[num] = -1;
	}

#endif
	return fd;
}


void	set_dev( int fd, int b )
{
struct	termios t;
int	baud[] = {B9600, B19200, B38400, B57600, B115200};

	if (tcgetattr(fd, &t) < 0) {
		printf("tcgetattr\n");
		exit(1);
	}
	
	t.c_cflag = CS8;
	t.c_cflag |= CSTOPB;
	t.c_cflag |= CREAD | HUPCL | CLOCAL;
	t.c_oflag = 0;
	t.c_iflag = IGNBRK | IGNPAR;
	t.c_lflag = 0;
		
	cfsetospeed(&t, baud[b] );
	cfsetispeed(&t, baud[b] );

	if (tcsetattr(fd, TCSANOW, &t) < 0) {
		printf("tcsetattr\n");
		exit(1);
	}

	return;
}	

void	show_write( int port, int sz )
{
int	offset = port * PORTROOM + GLOBLOFFSET+1;

	gotoxy( offset, 60 );
	printf("(%d bytes sent)", sz);
}

void	showframe( int port )
{
int	offset	= port * PORTROOM + GLOBLOFFSET;


	gotoxy( offset , 0);
	printf("%c[1;44mPort %d %c[0;0m%c[1;33m|%c[1;0m", ESC, port, ESC, ESC, ESC);

	gotoxy( offset++ , 52);
	printf("       %c[1;33m|%c[0m", ESC, ESC);

	gotoxy( offset , 0);
	printf("       0 %%");
	
	gotoxy( offset ,49);
	printf("     100 %%");
	
}

void	showbar(int port, int val, int all, char mark)
{
int	offset = port * PORTROOM + GLOBLOFFSET;
int	i;
	
	gotoxy( offset , 9);
	for(i=0;i<val*50/all && i < 50;i++) {
		printf("%c", mark);
	}	

	gotoxy( offset, 62);
	clrlin();
	if ( val >= all )
		printf("100 %%");
	else
		printf("%d %%", 100*val/all);
	
	fflush( stdout );
}


int main(int argc, char *argv[])
{
int	fd[MAXPORT];
int	n;
int	i,j;
int	ridx[MAXPORT], widx[MAXPORT];
int	iidx[MAXPORT];
int	numport;
int	start;
int	finish = 0;
int	ret;
int	bytes=TEST_SIZE;

fd_set	rfds, wfds;
char	rbuf[BUFSIZE];
char	wbuf[BUFSIZE];
time_t	t1,t2;
char	*ptr;
struct	timeval	tv;
int	b = 0;
char	*baudstr[] = { "9600", "19200", "38400", "57600", "115200" };
char	mark='=';
	
	if (argc < 3) {
		printf("Usage:%s dev num [baud] [bytes]\n\n", argv[0]);
		printf(" dev: the starting device.\n");
		printf(" num: the number ports following the starting device.\n");
		printf(" [baud]: The default baud is 9600.\n");
		printf("	 0 -> 9600\n");
		printf("	 1 -> 19200\n");
		printf("	 2 -> 38400\n");
		printf("	 3 -> 57600\n");
		printf("	 4 -> 115200\n");
		printf(" [bytes]: This is the total bytes to send. \n");
		printf("          If not specifing, the default value is 128 KB\n");
		exit(0);
	}

	numport = atoi(argv[2]);
	if (numport > MAXPORT) {
		printf("port number %d is not allowed. MAXIUM %d\n", numport, MAXPORT);
		exit(1);
	}

	for(ptr=argv[1]; *ptr; ptr++) {
		if (isdigit( *ptr ) )
			break;
	}
	
	start = atoi( ptr );
	printf("Start Port is %d\n", start);
	
	*ptr = '\0';
	
	if (argc >= 4) {
		b = atoi( argv[3] );
		if ( b < 0 || b > 4)
			b = 0;
	}

	if (argc >= 5) {
		bytes = atoi( argv[4] );
		if (bytes < 0)
			bytes = TEST_SIZE;
	}
	
	//
	//	Welcome msg
	//
	
	clrscr();
	gotoxy(0,0);
	
	printf("This loopback test is going to send %d bytes using serial port.\n", bytes);

	for(i=0;i<numport;i++) {
		fd[i] = open_dev( argv[1], start + i);
		set_dev( fd[i], b );

		showframe(i);

	}

	memset( ridx, 0, sizeof(ridx));
	memset( widx, 0, sizeof(widx));
	memset( iidx, 0, sizeof(iidx));

	for(i=0;i<CHUNK_SIZE;i++) {
		wbuf[i] = (char)i;
	}


	time(&t1);

	fflush( stdout );
	
	for(;;) {
		if (finish == numport)
			goto OUT;
			
		FD_ZERO( &rfds );
		FD_ZERO( &wfds );
		
		for(i=0;i<numport;i++) {
			if (fd[i] != -1) {
				FD_SET( fd[i], &rfds );
				FD_SET( fd[i], &wfds );	
			}
		}
		
		tv.tv_sec = 1;
		tv.tv_usec= 0;
		
		n = select( getdtablesize() , &rfds, &wfds, NULL, &tv);
		
		if (n < 0) {
			printf("Select ERROR !!\n");
			exit(1);
		}
		
		if (n == 0) {
			printf("Timeout !!\n");
			continue;
		}
		

		for(i=0;i<numport;i++) {
			mark = '=';
			if (fd[i] != -1 && FD_ISSET( fd[i], &rfds )) {
				n = read( fd[i], rbuf, BUFSIZE );
				for(j=0;j<n;j++, ridx[i]++) {
					if (rbuf[j] != (char)(ridx[i] % CHUNK_SIZE)) {
					mark = 'X';
					}
				}
				
				showbar(i, ridx[i], bytes, mark);
				
				if (ridx[i] > bytes) {
					time(&t2);
					
					gotoxy(i*PORTROOM+GLOBLOFFSET+1, 1);
					printf("Port %d takes %d seconds to transfer %d bytes\n", i, (int)(t2-t1), bytes);
					close(fd[i]);
					fd[i] = -1;
					finish++;
				}
			}
		}
		for(i=0;i<numport;i++) {
		
			if (fd[i] != -1 && FD_ISSET( fd[i], &wfds )) {
				if (widx[i] > bytes)
					continue;
				ret = write( fd[i], wbuf + iidx[i] , CHUNK_SIZE - iidx[i] );

#ifdef	LOG_FILE

				if (logfd[i] != -1)
					write( logfd[i], wbuf + iidx[i] , ret );

#endif
				iidx[i] += ret;
				iidx[i] %= CHUNK_SIZE;
				widx[i] += ret;
				show_write(i, widx[i] );
			}
		}
	}
OUT:
	gotoxy( 19, 0);
	printf("\nThe minimum time for each port using %s 8N1 is\n", baudstr[b] );
	printf(" %dx11/%s = ", bytes, baudstr[b] );
 	printf("%d\n", bytes*11/atoi(baudstr[b]));

	return 0;
}

